SystemOrganization addCategory: #DebuggerModel!

Object subclass: #DebugContext
	instanceVariableNames: 'context topContext'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'DebuggerModel'!

!DebugContext commentStamp: '<historical>' prior: 0!
A DebugContext is a helper that complements DebugSession. It is meant to be created dynamically on a context when the session wants to access the provided services.

To create instances first call forContext: to set the current context, and then if
the interrupted is different use topContext:

Not sure if it is still a good idea to have this class. 

Instance Variables
	context:		<Object>
	method:		<Object>
	methodNode:		<Object>
	ranges:		<Object>
	topContext:		<Object>

context
	- xxxxx

method
	- xxxxx

methodNode
	- xxxxx

ranges
	- xxxxx

topContext
	- xxxxx
!

!DebugContext class methodsFor: 'instance creation' stamp: 'AndreiChis 1/7/2013 10:29'!
forContext: aContext 
	^ self new forContext: aContext ! !

!DebugContext methodsFor: 'private' stamp: 'AndreiChis 1/14/2013 16:23'!
blockNotFoundDialog: aMethod with: aText

	| browser message result |
	
	message := 'Method for block not found on stack, can''t edit and continue'.
	
	"shouldn't edit doits"
	aMethod selector isDoIt
		ifTrue: [ 	^ self inform: message ].
			
	result := UIManager default
						confirm: message
						trueChoice: 'Browse' translated
						falseChoice: 'Cancel' translated.
						
	"possible return values are true | false | nil"
	result == true ifFalse: [ ^ self ].
		
	"let's browse the given method with the edited contents"
	browser := aMethod browse.
	browser 
		contents: aText; 
		changed: #contents.
	browser codeTextMorph 
		setSelection: self codeTextMorph selectionInterval;
		hasUnacceptedEdits: true! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/11/2013 17:51'!
context
	^ context! !

!DebugContext methodsFor: 'evaluating actions' stamp: 'AndreiChis 1/7/2013 10:29'!
evaluate: expression
	^ Compiler new
		evaluate: expression
		in: context
		to: context receiver! !

!DebugContext methodsFor: 'initialization' stamp: 'AndreiChis 2/6/2013 14:35'!
forContext: aContext 
	context := aContext.
	topContext := aContext.

	"methodNode := method methodNode.
	ranges := Dictionary new."
	" Put all ranges with a possitive Key PC in the ranges dictionary "
	"methodNode rawSourceRanges
		keysAndValuesDo: [ :node :range |
			node pc = 0
				ifFalse: [ ranges at: node pc put: range ]]."


	" Sort all the Key PCs in the sortedKeyPCs "
	"sortedKeyPCs := SortedCollection new."
	" ranges keysDo: [ :pc | sortedKeyPCs add: pc ]."! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
gtDebuggerRetrieveVariableValuePairs
	|fieldList|

	fieldList := OrderedCollection new.
	fieldList 
		add: ('_self' ->  context receiver);
	      addAll: (context receiver class allInstVarNames collect: [:each | each -> (context receiver instVarNamed: each) ]).
	
	fieldList add: ('_thisContext' ->  context).
	 context == nil ifTrue: [^fieldList].
	fieldList 
		add: ('_stack top' -> ( context stackPtr > 0 ifTrue: [ context top])).
		
	(context tempNames size > 0) ifTrue: [
		fieldList addAll: ((1 to: context tempNames size) collect: [ :index || key value |
			key := context tempNames at: index.
			value := [context debuggerMap namedTempAt: index in: context] 
				on: Error 
				do: [value := (Text string: 'error obtaining field value' attribute: TextColor red) asString ].
			key -> value ] ) ].
	
	^ fieldList! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/14/2013 16:26'!
locateClosureHome
	"In case the current context is a BlockContext locate the closureHome and ask the
	user to validate the new context. If closureHome is not found or the user does not 
	validate the new context, return nil. 
	If the current context is not a BlockContext return it."
	| closureHome result aText |
	
	context isExecutingBlock ifTrue: [
		closureHome := context activeHome.
		closureHome ifNil: [
			self blockNotFoundDialog: context method with: aText.
		 	^ nil ].
		 (self confirm: 'I will have to revert to the method from\which this block originated.  Is that OK?' withCRs) 
			ifFalse: [ ^ nil ].
		^ closureHome].
	
	^ context! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
receiver
	^ context receiver! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
receiverClass
	"Answer the class of the receiver. 
	It may differ from 'self selectedClass' "
	
	^ context receiver class! !

!DebugContext methodsFor: 'evaluating actions' stamp: 'AndreiChis 1/7/2013 10:29'!
recompileCurrentMethodTo: aText 
	| classOfMethod selector |
	
	classOfMethod := self selectedClass.
	selector := classOfMethod parserClass new parseSelector: aText.
	(selector == self selectedMessageName or: [(self selectedMessageName beginsWith: 'DoIt')
		and: [selector numArgs = self selectedMessageName numArgs]]) ifFalse: [
			self inform: 'can''t change selector'. ^ nil].
		
	selector := classOfMethod 
		compile: aText 
		classified: self selectedMessageCategoryName 
		notifying: nil.
	selector isNil 
		ifFalse: [ ^ classOfMethod compiledMethodAt: selector. ]
		ifTrue: [ ^ nil ] 
		! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
selectedClass
	"Answer the class in which the current context's method was found."

	^ context methodClass! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
selectedMessageCategoryName
	"Answer the name of the message category of the message of the current context."

	^ self selectedClass organization categoryOfElement: self selectedMessageName! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
selectedMessageName
	"Answer the message selector of the current context.
	 If the method is unbound we can still usefully answer its old selector."

	| selector |
	selector := context methodSelector.
	^(selector ~~ context method selector
	    and: [selector beginsWith: 'DoIt'])
		ifTrue: [context method selector]
		ifFalse: [selector]! !

!DebugContext methodsFor: 'private' stamp: 'AndreiChis 1/7/2013 12:03'!
source
	"Answer the source code of the currently selected context."
	^ context sourceCode

	"Old implementation"
	"^ method holdsTempNames
		ifTrue: [ method getSource ]
		ifFalse: [ methodNode sourceText ]"! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 12:03'!
sourceText
	^ self source asText makeSelectorBold! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
topContext: aContext
	topContext  := aContext ! !

Object subclass: #DebugSession
	instanceVariableNames: 'context process errorWasInUIProcess'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'DebuggerModel'!

!DebugSession commentStamp: '<historical>' prior: 0!
A DebugSession models a debuggeing session. It contains the interrupted context and process. Its main goal is to handle debugger actions such as restart or stepInto, as well as recomplilation of methods. It is the model used as an input to a ui.

As it is just a model it does now contain any information related to the ui. For example, it does not know what a selection in the ui is. It is the job on the ui to maintain the selection and call this session with the propper context.

To create sessions use the mehod 'process: aProcess context: aContext'. aContext must be a context belonging to aProcess, and aProcess must be an interrupted process.

Instance Variables
	context:		<Object>
	errorWasInUIProcess:		<Object>
	process:		<Object>

context
	- xxxxx

errorWasInUIProcess
	- xxxxx

process
	- xxxxx
!

!DebugSession class methodsFor: 'instance creation' stamp: 'AndreiChis 1/7/2013 10:29'!
process: aProcess context: aContext
	^ self new
		process: aProcess context: aContext! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/7/2013 10:29'!
activePC: aContext
	^ (self isLatestContext: aContext)
		ifTrue: [ context pc ]
		ifFalse: [ self previousPC: aContext ].! !

!DebugSession methodsFor: 'private evaluating' stamp: 'AndreiChis 1/14/2013 15:25'!
askForCategoryIn: aClass default: aString
	| categories category |
	categories := OrderedCollection withAll: (aClass allMethodCategoriesIntegratedThrough: Object).	
	category := UIManager default  
		chooseOrRequestFrom: categories
		title: 'Please provide a good category for the new method!!' translated.
	category ifNil: [^aString].
	^ category isEmptyOrNil 
		ifTrue: [^ aString] 
		ifFalse: [ category ]! !

!DebugSession methodsFor: 'private evaluating' stamp: 'AndreiChis 1/14/2013 15:20'!
askForSuperclassOf: aClass toImplement: aSelector ifCancel: cancelBlock
	| classes chosenClassIndex |
	classes := aClass withAllSuperclasses.
	chosenClassIndex := UIManager default 
		chooseFrom: (classes collect: [:c | c name])
		title: 'Define #', aSelector, ' in which class?'.
	chosenClassIndex = 0 ifTrue: [^ cancelBlock value].
	^ classes at: chosenClassIndex! !

!DebugSession methodsFor: 'private' stamp: 'AndreiChis 1/15/2013 11:22'!
clear
	process := nil.
	self updateContextTo: nil! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
context
	^ context! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/31/2013 14:41'!
contextChanged

	self triggerEvent: #contextChanged! !

!DebugSession methodsFor: 'evaluating actions' stamp: 'AndreiChis 1/14/2013 13:36'!
createMethodInContext: selectedContext
	"Should only be called when the debugger was created in response to a
	MessageNotUnderstood exception. Create a stub for the method that was
	missing and proceed into it."
	
	| msg chosenClass |

	"Added due to error 3011 - Hernan"
	selectedContext = self context ifFalse: [ ^ UIManager default inform: 'Please select the #doesNotUnderstand: context'].
	
	msg := self context tempAt: 1.
	chosenClass := self
		askForSuperclassOf: self context receiver class
		toImplement: msg selector
		ifCancel: [^self].
	self implement: msg inClass: chosenClass.! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:40'!
createModelForContext: aContext
	^ (DebugContext forContext: aContext) topContext: context! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 2/6/2013 14:15'!
downInContext: aContext
	"move down the context stack to the previous (enclosing) context"

	self flag: 'This does not take into account (bypasses) filtering'.
	^ aContext sender. ! !

!DebugSession methodsFor: 'initialization' stamp: 'AndreiChis 1/15/2013 15:45'!
errorWasInUIProcess: aBoolean
	errorWasInUIProcess := aBoolean! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/7/2013 10:29'!
filterTopContext: aContext
	|ctx|
	 
	ctx := aContext.
	(self shouldDisplayOnTopContext: ctx) ifTrue: [^ ctx].
	[(ctx := ctx sender) ~~ nil] whileTrue:
		[(self shouldDisplayOnTopContext: ctx) ifTrue: [^ ctx]].
	^ nil! !

!DebugSession methodsFor: 'private evaluating' stamp: 'AndreiChis 1/31/2013 14:15'!
implement: aMessage inClass: aClass
	
	aClass
		compile: (DynamicMessageImplementor for: aMessage in: aClass) value
		classified: (self askForCategoryIn: aClass default: 'as yet unclassified').
	
	self context privRefreshWith: (aClass lookupSelector: aMessage selector).
	self context method numArgs > 0 ifTrue:
		[(self context tempAt: 1) arguments withIndexDo:
			[:arg :index|
			self context tempAt: index put: arg]].
	
	self contextChanged
! !

!DebugSession methodsFor: 'testing' stamp: 'AndreiChis 1/14/2013 18:01'!
isContextPostMortem: selectedContext
	"return whether we're inspecting a frozen exception without a process attached"
	| suspendedContext |
	suspendedContext := process suspendedContext.
	suspendedContext ifNil: [ ^ false ].
	(suspendedContext == selectedContext)
		ifTrue: [ ^ false ].
	^ (suspendedContext findContextSuchThat: [:c | c sender == selectedContext]) isNil! !

!DebugSession methodsFor: 'testing' stamp: 'AndreiChis 1/7/2013 11:21'!
isInterruptedContextATest

	^ (self isTestObject: self context receiver) and: [ self isTestMethod: self context method of: self context receiver ]! !

!DebugSession methodsFor: 'testing' stamp: 'AndreiChis 1/14/2013 13:23'!
isInterruptedContextDoesNotUnderstand	
	
	^ self context selector == #doesNotUnderstand:! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/7/2013 10:29'!
isLatestContext: aContext
	^ process suspendedContext == aContext! !

!DebugSession methodsFor: 'testing' stamp: 'AndreiChis 1/7/2013 11:17'!
isTestMethod: aCompiledMethod of: aTestCase

	^ aCompiledMethod selector = aTestCase selector! !

!DebugSession methodsFor: 'testing' stamp: 'AndreiChis 1/7/2013 11:17'!
isTestObject: anObject

	"I'm not sure this is the best way to doit because it creates a coupling with TestCase, but due that SUnit is part of the core I think it is not bad after all - Hernan'"
	^ anObject isKindOf: TestCase! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/13/2013 19:19'!
pcRangeForContext: aContext
	"Answer the indices in the source code for the method corresponding to 
	aContext's program counter value."

	aContext isDead ifTrue:
		[^1 to: 0].
	^aContext debuggerMap
		rangeForPC: aContext pc
		contextIsActiveContext: (self isLatestContext: aContext)! !

!DebugSession methodsFor: 'private' stamp: 'AndreiChis 1/7/2013 11:21'!
prepareTestToRunAgain

	self context receiver prepareToRunAgain! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/7/2013 10:29'!
previousPC: aContext
	^ (aContext method pcPreviousTo: aContext pc)
		ifNil: [ aContext pc ]! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
process
	^ process! !

!DebugSession methodsFor: 'initialization' stamp: 'AndreiChis 1/7/2013 10:29'!
process: aProcess context: aContext
	aProcess stepToSendOrReturn.
	process := aProcess.
	context := self filterTopContext: aContext.
	context isNil ifTrue: [context := aContext].
	
! !

!DebugSession methodsFor: 'evaluating actions' stamp: 'AndreiChis 2/6/2013 14:22'!
recompileMethodTo: text inContext: selectedContext
	"The retrieved information has changed and its source must now be updated.
	 In this case, the retrieved information is the method of the given context."
	| newMethod recompilationContext |				
				
	(recompilationContext := (self createModelForContext: selectedContext) locateClosureHome) 
		ifNil: [ ^ nil ]. 
				
	newMethod := (self createModelForContext: recompilationContext) recompileCurrentMethodTo: text.
	newMethod ifNil: [ ^ nil ].
	newMethod isQuick ifTrue: [
		recompilationContext := self downInContext: recompilationContext.
		recompilationContext jump: (recompilationContext previousPc - recompilationContext pc) ].
		
	self rewindContextToMethod: newMethod fromContext: recompilationContext. 
	
	self contextChanged.
	^ newMethod! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/31/2013 15:06'!
restart: selectedContext
	"Proceed from the initial state of selectedContext." 
	"Closing now depends on a setting (RestartAlsoProceeds class variable) --> not supported in this version"

	(self isContextPostMortem: selectedContext) ifTrue: [^ self].
	self unwindAndRestartToContext: selectedContext.
	
	"Issue 3015 - Hernan"
	self isInterruptedContextATest ifTrue: [ self prepareTestToRunAgain ].
	
	self triggerEvent: #restart		
	! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/15/2013 15:33'!
resume
	"Proceed execution of the receiver's model, starting after the expression at 
	which an interruption occurred."
	"If the user of this session does not plan to reuse it, it should call 'self clean' "
	
	Smalltalk okayToProceedEvenIfSpaceIsLow ifTrue: [
		(self isContextPostMortem: self context) ifTrue: [^ self].
		self resumeProcess.
	
		self triggerEvent: #resume].
	! !

!DebugSession methodsFor: 'private' stamp: 'AndreiChis 1/15/2013 15:41'!
resumeProcess
	"Make sure the interrupted process is restored properly and restart the low space handler"
	
	process isTerminated
		ifFalse: [errorWasInUIProcess
				ifTrue: [UIManager default resumeUIProcess: process]
				ifFalse: [process resume]].

	Smalltalk installLowSpaceWatcher.	"restart low space handler".
		
	! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/31/2013 15:19'!
rewindContextToMethod: aMethod fromContext: aContext
	"this method is typically to be used after a hot compilation of a method from the stack.
	in order to return to the context containg the compiled method."
	| ctxt |
	ctxt :=  process popTo: aContext.
	
	ctxt == aContext
		ifFalse: [
			self inform: 'Method saved, but current context unchanged\because of unwind error. Click OK to see error' withCRs ]
		ifTrue: [
			aMethod isQuick ifFalse: [
				process
					restartTopWith: aMethod;
					stepToSendOrReturn ] ].
	self updateContextTo:  ctxt.
	
	"Issue 3015 - Hernan"
	self isInterruptedContextATest ifTrue: [ self prepareTestToRunAgain ].! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/31/2013 15:13'!
runToSelection: selectionInterval inContext: selectedContext 
	"Attempt to step over instructions in selectedContext until the 
	execution reaches the selected instruction. This happens when the
	program counter passes the begining of selectionInterval.
	
	A not nill and valid interval is expected." 
	
	(self pcRangeForContext: selectedContext) first >= selectionInterval first 
		ifTrue: [ ^self ].
	self stepOver: selectedContext.
	[ selectedContext == self context and: [ (self pcRangeForContext: selectedContext) first < selectionInterval first ] ] 
		whileTrue: [ self stepOver: selectedContext ]! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 2/6/2013 14:33'!
selectedCodeRangeForContext: selectedContext

	^ self pcRangeForContext: selectedContext! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/7/2013 10:29'!
shouldDisplayOnTopContext: aContext
	
	^ aContext method selector ~= #halt! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
stack
	^ context stack! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
stackOfSize: size
	^ context stackOfSize: size! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 10:29'!
stepInto
	self stepInto: context! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/15/2013 11:17'!
stepInto: selectedContext
	"Send the selected message in selectedContext, and take control in 
	the method invoked to allow further step or send."

	(self isContextPostMortem: self context) ifTrue: [^ self].

	process step: selectedContext.
	self updateContextTo:  process stepToSendOrReturn.
	
	self triggerEvent: #stepInto! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 10:29'!
stepOver
	self stepOver: context! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/15/2013 11:18'!
stepOver: selectedContext
	"Send the selected message in selectedContext, and regain control 
	after the invoked method returns."
	
	| newContext |
	
	(self isContextPostMortem: self context) ifTrue: [^ self].
	
	newContext := process completeStep: selectedContext.
	self updateContextTo:  
		(newContext == selectedContext
			ifTrue: [ process stepToSendOrReturn ]
			ifFalse: [ newContext ]).

	self triggerEvent: #stepOver ! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 10:29'!
stepThrough
	self stepThrough: context! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/15/2013 11:18'!
stepThrough: selectedContext
	"Send messages until you return to selectedContext.
	 Used to step into a block in the method."
	
	(self isContextPostMortem: self context) ifTrue: [^ self].
	
	process stepToHome: selectedContext.
	self updateContextTo:  process stepToSendOrReturn.
			
	self triggerEvent: #stepThrough ! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/15/2013 15:36'!
terminate
	"Action that needs to be executed after the window containing this debug session is closed, 
	in order to terminate the right process."
	
	self process isNil 
		ifFalse: [ 
			"Assume the user closed the debugger. Simply kill the interrupted process."
			self process terminate.
			self clear.
			Smalltalk installLowSpaceWatcher.  "restart low space handler" ]
		ifTrue: [ 
			"Assume the interrupted process was resumed."
			"Kill the active process if the error was in the UI as there should be only one UI process."
			errorWasInUIProcess == false
				ifFalse: [Processor terminateActive] ]
	! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/31/2013 15:21'!
unwindAndRestartToContext: aContext
	|ctx|

	ctx := process popTo: aContext.
	ctx == aContext ifTrue: [ 
		"Only restart the process if the stack was unwind"
		process
			restartTop;
			stepToSendOrReturn ].
	self flag: 'Should a warning be displayed if the the unwind failed?'.
	self updateContextTo:  aContext! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/15/2013 11:15'!
updateContextTo: aContext

	context := aContext! !

!DebugSession methodsFor: 'updating' stamp: 'AndreiChis 1/7/2013 10:29'!
updateWithContext: newContext fromProcess: aProcess
	self process: aProcess context: newContext
	 
	"self process: aProcess.
	self unwindTo: (self filterTopContextAndUnwindForContext: newContext).
	self process stepToSendOrReturn."! !
